Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var ix_1 = require("ix");
var rxjs_1 = require("rxjs");
var Compare_1 = require("../../../Utils/Compare");
var ListItemsViewModel_1 = require("../ListItems/ListItemsViewModel");
var PagerViewModel_1 = require("../Pager/PagerViewModel");
var DataGridViewModel = /** @class */ (function (_super) {
    tslib_1.__extends(DataGridViewModel, _super);
    /**
     * @param source data source.
     * @param pager pager. if omitted a default pager will be created. use null for no pager.
     * @param context request context included in projection requests. if included requests are bound to context events.
     * @param comparer custom object comparer. if omitted a default object comparer will be used.
     */
    function DataGridViewModel(source, pager, context, comparer, rateLimit) {
        if (comparer === void 0) { comparer = new Compare_1.ObjectComparer(); }
        if (rateLimit === void 0) { rateLimit = DataGridViewModel.DefaultRateLimit; }
        var _this = _super.call(this, source) || this;
        _this.pager = pager === null ? null : pager || new PagerViewModel_1.PagerViewModel();
        _this.comparer = String.isString(comparer)
            ? new Compare_1.ObjectComparer(comparer)
            : comparer;
        _this.processRequests = _this.wx.command();
        _this.sort = _this.wx.command();
        _this.toggleSortDirection = _this.wx.command();
        _this.sorting = _this.wx.whenAny(_this.sort, function (x) { return x; }).toProperty();
        _this.requests = _this.getRequests(context)
            // combineLatest will "gate" the requests until the processRequests command is invoked
            .combineLatest(_this.processRequests.results.take(1), function (x) { return x; })
            .toProperty(undefined, false);
        _this.responses = _this.getResponses(_this.requests, rateLimit).toProperty(undefined, false);
        _this.isLoading = rxjs_1.Observable.merge(_this.requests.changed.map(function () { return true; }), _this.responses.changed.map(function () { return false; })).toProperty(true);
        var validResponses = _this.wx
            .whenAny(_this.responses, function (x) { return x; })
            .filterNull()
            .share();
        _this.projectedSource = validResponses
            .map(function (x) { return x.items; })
            .toProperty(ix_1.Iterable.empty(), false);
        _this.projectedCount = validResponses.map(function (x) { return x.count; }).toProperty();
        if (_this.pager != null) {
            _this.addSubscription(_this.wx
                .whenAny(_this.projectedCount, function (x) { return x; })
                .filterNull()
                .invokeCommand(_this.pager.updateCount));
            _this.addSubscription(_this.wx
                .whenAny(_this.sort, function () { return 1; })
                .invokeCommand(_this.pager.selectPage));
        }
        _this.addSubscription(_this.wx
            .whenAny(_this.toggleSortDirection, function (x) { return x; })
            .withLatestFrom(_this.wx.whenAny(_this.requests, function (x) { return x; }), function (field, request) { return ({ field: field, request: request }); })
            .map(function (x) {
            var sortArgs = {
                field: x.field,
                direction: _this.getReverseSortDirection(x.field, x.request),
            };
            return sortArgs;
        })
            .invokeCommand(_this.sort));
        _this.addSubscription(_this.wx
            .whenAny(_this.sort, function (x) { return _this.selectedItems.value; })
            .invokeCommand(_this.selectItems));
        _this.addSubscription(_this.wx.whenAny(_this.sort, function (x) { return x; }).subscribe(function (x) {
            _this.notifyChanged(x);
        }));
        return _this;
    }
    DataGridViewModel.prototype.isRoutingStateHandler = function () {
        return true;
    };
    DataGridViewModel.prototype.createRoutingState = function (changed) {
        return Object.trim({
            pager: this.getRoutingStateValue(this.pager, function (x) {
                return x.createRoutingState(changed);
            }),
            sorting: this.getRoutingStateValue(this.sorting.value),
        });
    };
    DataGridViewModel.prototype.applyRoutingState = function (state) {
        if (this.pager != null) {
            this.pager.applyRoutingState(state.pager || {});
        }
        if (this.isSortChanged(state.sorting)) {
            this.sort.execute(state.sorting);
        }
        // notify our streams that we're routed and may begin processing requests
        this.processRequests.execute();
    };
    DataGridViewModel.prototype.getItemsSourceProperty = function () {
        return this.projectedSource || _super.prototype.getItemsSourceProperty.call(this);
    };
    DataGridViewModel.prototype.isSortChanged = function (sort) {
        var sorting = this.sorting.value;
        if (sorting == null && sorting == null) {
            return false;
        }
        if (sort == null || sorting == null) {
            return true;
        }
        return sort.direction !== sorting.direction || sort.field !== sorting.field;
    };
    DataGridViewModel.prototype.getReverseSortDirection = function (field, request) {
        // we have no sort direction state
        if (request == null ||
            request.sort == null ||
            request.sort.direction == null) {
            return Compare_1.SortDirection.Ascending;
        }
        // we have no sort field state or we are sorting a new field
        if (request.sort.field == null || request.sort.field !== field) {
            return Compare_1.SortDirection.Ascending;
        }
        // reverse the current sort direction
        return request.sort.direction === Compare_1.SortDirection.Ascending
            ? Compare_1.SortDirection.Descending
            : Compare_1.SortDirection.Ascending;
    };
    DataGridViewModel.prototype.getRequest = function (source, page, sort, context) {
        return {
            page: page,
            sort: sort,
            context: context,
        };
    };
    DataGridViewModel.prototype.getRequests = function (context) {
        var _this = this;
        var pagerObservable = this.pager == null ? rxjs_1.Observable.of(undefined) : this.pager.requests;
        var source = this.wx
            .whenAny(this.source, function (x) { return x; })
            .filter(function (x) { return x !== _this.emptySource; });
        return this.wx.whenAny(source, pagerObservable, this.sorting, context || rxjs_1.Observable.of(undefined), function (src, page, sort, ctx) {
            return _this.getRequest(src, page, sort, ctx);
        });
    };
    DataGridViewModel.prototype.getResponse = function (request) {
        if (request == null) {
            return undefined;
        }
        var items = ix_1.Iterable.from(this.source.value);
        return this.getResponseFromItems(items, request);
    };
    DataGridViewModel.prototype.getResponseFromItems = function (items, request) {
        var count = items.count();
        if (this.comparer != null &&
            request.sort != null &&
            !String.isNullOrEmpty(request.sort.field) &&
            request.sort.direction != null) {
            items = this.comparer.sortIterable(items, request.sort.field, request.sort.direction);
        }
        if (request.page != null) {
            if ((request.page.offset || 0) > 0) {
                items = items.skip(request.page.offset);
            }
            if ((request.page.limit || 0) > 0) {
                items = items.take(request.page.limit);
            }
        }
        return {
            items: items,
            count: count,
        };
    };
    DataGridViewModel.prototype.getResponses = function (requests, rateLimit) {
        var _this = this;
        if (rateLimit === void 0) { rateLimit = DataGridViewModel.DefaultRateLimit; }
        return (this.wx
            .whenAny(requests || this.requests, function (x) { return x; })
            // because requests can be injected here, we cannot trust that nulls are not already filtered out
            // so we filter here just to be safe (this should be a no-op in most cases)
            .filterNull()
            .debounceTime(rateLimit)
            .flatMap(function (x) {
            return rxjs_1.Observable.defer(function () {
                return _this.wx.getObservable(_this.getResponse(x));
            }).catch(function (e) {
                _this.alertForError(e, 'Data Response Error');
                return _this.wx.getObservable({
                    items: ix_1.Iterable.empty(),
                    count: 0,
                });
            });
        })
            .filterNull());
    };
    DataGridViewModel.displayName = 'DataGridViewModel';
    DataGridViewModel.DefaultRateLimit = 100;
    return DataGridViewModel;
}(ListItemsViewModel_1.ListItemsViewModel));
exports.DataGridViewModel = DataGridViewModel;
//# sourceMappingURL=DataGridViewModel.js.map