Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var rxjs_1 = require("rxjs");
var DataGridViewModel_1 = require("../DataGrid/DataGridViewModel");
var SearchViewModel_1 = require("../Search/SearchViewModel");
var ItemListPanelViewModel = /** @class */ (function (_super) {
    tslib_1.__extends(ItemListPanelViewModel, _super);
    /**
     * @param source data source.
     * @param filterer filter predicate. executed for each item when the search context is available.
     * @param search search handler. if omitted a default search handler will be created. use null for no search handling.
     * @param pager pager. if omitted a default pager will be created. use null for no pager.
     * @param context request context included in projection requests. if included requests are bound to context events.
     * @param comparer custom object comparer. if omitted a default object comparer will be used.
     */
    function ItemListPanelViewModel(source, filterer, search, pager, context, comparer, rateLimit) {
        var _this = this;
        // create the effective search view model (or null if omitted)
        search = ItemListPanelViewModel.getItemListPanelSearch(search);
        // initialize the data grid with a composite context observable
        _this = _super.call(this, source, pager, ItemListPanelViewModel.getDataGridContext(search, context), comparer, rateLimit) || this;
        _this.filterer = filterer;
        _this.search = search;
        if (_this.search != null) {
            // seed the search after routing state is loaded
            // we need to do this because we want an initial request even if the filter is empty
            _this.processRequests.results.take(1).invokeCommand(_this.search.search);
        }
        return _this;
    }
    ItemListPanelViewModel.getItemListPanelSearch = function (search) {
        return search === null ? null : search || new SearchViewModel_1.SearchViewModel();
    };
    ItemListPanelViewModel.createItemListPanelContext = function (search) {
        return {
            search: search,
        };
    };
    ItemListPanelViewModel.getItemListPanelContext = function (search) {
        if (search == null) {
            return rxjs_1.Observable.of({});
        }
        return this.wx
            .whenAny(search.requests, function (x) { return x; })
            .filterNull()
            .map(function (x) { return ItemListPanelViewModel.createItemListPanelContext(x); });
    };
    ItemListPanelViewModel.getDataGridContext = function (search, context) {
        var itemListPanelContext = ItemListPanelViewModel.getItemListPanelContext(search);
        if (context == null) {
            // no custom context supplied so just use the standard item list panel context
            return itemListPanelContext;
        }
        return this.wx.whenAny(itemListPanelContext, context, function (ilpCtx, userCtx) {
            return Object.assign({}, ilpCtx, userCtx);
        });
    };
    ItemListPanelViewModel.getSearchRequest = function (request) {
        if (request.context == null ||
            request.context.search == null ||
            request.context.search.regex == null) {
            return undefined;
        }
        return request.context.search;
    };
    ItemListPanelViewModel.filterItems = function (items, searchRequest, filterer) {
        return items.filter(function (x) { return filterer(x, searchRequest); });
    };
    ItemListPanelViewModel.prototype.isRoutingStateHandler = function () {
        return true;
    };
    ItemListPanelViewModel.prototype.createRoutingState = function (changed) {
        return Object.trim(Object.assign(_super.prototype.createRoutingState.call(this, changed), {
            search: this.getRoutingStateValue(this.search, function (x) {
                return x.createRoutingState(changed);
            }),
        }));
    };
    ItemListPanelViewModel.prototype.applyRoutingState = function (state) {
        if (this.search != null) {
            this.search.applyRoutingState(state.search || {});
        }
        // we apply the base (grid) routing state last because it will trigger requests to start
        _super.prototype.applyRoutingState.call(this, state);
    };
    ItemListPanelViewModel.prototype.getResponseFromItems = function (items, request) {
        // if no filterer was defined then just return default response
        if (this.filterer == null) {
            return _super.prototype.getResponseFromItems.call(this, items, request);
        }
        // try and extract the search request
        var searchRequest = ItemListPanelViewModel.getSearchRequest(request);
        // if no search request was involved then return the default response
        if (searchRequest == null) {
            return _super.prototype.getResponseFromItems.call(this, items, request);
        }
        // filter the items according to the search request involved
        var filteredItems = this.getFilteredItems(items, searchRequest);
        // generate a response from the filtered items
        return _super.prototype.getResponseFromItems.call(this, filteredItems, request);
    };
    ItemListPanelViewModel.prototype.getFilteredItems = function (items, searchRequest) {
        return ItemListPanelViewModel.filterItems(items, searchRequest, this.filterer);
    };
    ItemListPanelViewModel.displayName = 'ItemListPanelViewModel';
    return ItemListPanelViewModel;
}(DataGridViewModel_1.DataGridViewModel));
exports.ItemListPanelViewModel = ItemListPanelViewModel;
//# sourceMappingURL=ItemListPanelViewModel.js.map