Object.defineProperty(exports, "__esModule", { value: true });
var param = require("jquery-param");
var rxjs_1 = require("rxjs");
var Routing_1 = require("../Routing");
var Interfaces_1 = require("./Interfaces");
function sanitizeUri(uri) {
    // first check if our provided uri is already prepped for params
    if (uri.indexOf('?') >= 0) {
        // it has been prepped, so check to see if the last character is an '&'
        if (uri[uri.length - 1] !== '&') {
            // no '&' at the end, so append one
            return uri + '&';
        }
    }
    else {
        // it hasn't been prepped so just append a '?'
        return uri + '?';
    }
    // uri is already sane, just return it
    return uri;
}
exports.sanitizeUri = sanitizeUri;
function getUriFromParams(uri, params) {
    // first filter out any empty/null params
    params = Object.trim(params);
    // sanitize the uri
    uri = sanitizeUri(uri);
    // if we have parameters to append then append the encoded params
    if (params != null && Object.keys(params).length > 0) {
        return uri + param(params);
    }
    // otherwise return the sanitized uri as is
    return uri;
}
exports.getUriFromParams = getUriFromParams;
function getError(xhr, uri, logger) {
    if (xhr == null) {
        return {
            message: 'Invalid XMLHttpRequest (null)',
            uri: uri,
        };
    }
    var code = xhr.status > 0 ? xhr.status : undefined;
    var reason = String.isNullOrEmpty(xhr.statusText)
        ? undefined
        : xhr.statusText;
    var response = String.isNullOrEmpty(xhr.response)
        ? undefined
        : xhr.response;
    var message;
    var messageDetail;
    if (xhr.readyState === 4 && xhr.status === 0) {
        // this is the best heuristic we have for detecting a timeout
        // see: http://stackoverflow.com/a/38097577/2789877
        // NOTE: timeouts can be tested by supplying options param, with an
        // xmlHttpRequest creation function that injects timeout = 1, then
        // use the following as the uri:
        // https://httpbin.org/delay/5
        var windowLocation = Routing_1.getWindowLocation();
        if (windowLocation != null && uri.indexOf(windowLocation.origin) !== 0) {
            // NOTE: it is possible that you can reach this path if you attempt a CORS
            //       fetch where the OPTIONS preflight request 404's.
            // see: https://developer.mozilla.org/en-US/docs/Web/HTTP/Access_control_CORS#Preflighted_requests
            message = 'Request Timeout or Invalid Cross Origin Request';
        }
        else {
            message = 'Request Timeout';
        }
    }
    else if (response != null) {
        // something came back in the response, so let's try and extract an error
        if (!String.isNullOrEmpty(xhr.responseURL)) {
            // the responseURL will be more trustworth than our passed in parameter
            uri = xhr.responseURL;
        }
        var responseObject = void 0;
        if (String.isString(response)) {
            try {
                // try and parse the response as JSON
                responseObject = JSON.parse(response);
            }
            catch (e) {
                logger.debug('Unable to parse response', response, e);
                // JSON parsing didn't work, fallback on straight assignment
                message = response;
            }
        }
        else if (Object.isObject(response)) {
            responseObject = response;
        }
        if (responseObject != null) {
            // try and get the message and message detail
            message = responseObject.message || responseObject.Message;
            messageDetail =
                responseObject.messageDetail || responseObject.MessageDetail;
        }
        if (message == null && messageDetail == null) {
            // unable to produce a message yet, so fallback on stringify
            try {
                message = String.stringify(response, null, 2);
            }
            catch (e) {
                logger.warn('Unable to stringify response', response, e);
                // last ditch effort, just call toString on the object
                message = response.toString();
            }
        }
    }
    else {
        // we can't detect what type of error this is, so log the xhr and return a generic message
        logger.error('Invalid XHR Error', xhr);
        message = 'Invalid Response from API Host';
    }
    // message should always be non-null by this point, but just in case have a valid fallback
    if (message == null) {
        logger.error('Unable to extract error message', xhr);
        message = 'Unknown Error';
    }
    return {
        uri: uri,
        message: message,
        messageDetail: messageDetail,
        code: code,
        reason: reason,
        response: response,
    };
}
exports.getError = getError;
exports.defaultHeaders = {
    Accept: 'application/json',
    'Content-Type': 'application/json',
};
function getRequest(action, url, logger, method, params, data, options) {
    if (method === void 0) { method = Interfaces_1.HttpRequestMethod.GET; }
    logger.debug("getRequest: [" + method + "] " + action, {
        url: url,
        params: params,
        data: data,
        options: options,
    });
    url = getUriFromParams(url, params);
    logger.info("API Request: " + action + " (" + url + ")", data);
    var body = data == null ? undefined : String.stringify(data, null, 2);
    options = Object.assign({
        headers: exports.defaultHeaders,
        async: true,
        body: body,
        method: Interfaces_1.HttpRequestMethod[method],
        url: url,
    }, options);
    return rxjs_1.Observable.ajax(options)
        .map(function (x) { return x.response; })
        .do(function (x) {
        logger.info("API  Result: " + action + " (" + url + ")", x);
    })
        .catch(function (x) {
        logger.error("API  ERROR: " + action + " (" + url + ")", x);
        return rxjs_1.Observable.throw(getError(x.xhr, url, logger));
    });
}
exports.getRequest = getRequest;
function isStoreApi(value) {
    if (value == null) {
        return false;
    }
    var store = value;
    return (String.isString(store.path) &&
        String.isString(store.base) &&
        String.isString(store.baseUri) &&
        store.observe instanceof Function &&
        store.getObservable instanceof Function &&
        store.postObservable instanceof Function &&
        store.getSampleStoreValue instanceof Function);
}
exports.isStoreApi = isStoreApi;
//# sourceMappingURL=Helpers.js.map