// tslint:disable:max-classes-per-file
Object.defineProperty(exports, "__esModule", { value: true });
var moment = require("moment");
function isComparable(obj) {
    return obj.compareTo instanceof Function;
}
exports.isComparable = isComparable;
var ValueComparer = /** @class */ (function () {
    function ValueComparer(comparison) {
        if (comparison === void 0) { comparison = ValueComparer.DefaultComparison; }
        this.comparison = comparison;
    }
    ValueComparer.DefaultComparison = function (a, b) {
        if (a === b || (a == null && b == null)) {
            // both are null or the same, so equality is zero
            return 0;
        }
        else if (a == null || b == null) {
            // only one is null, non-null takes higher value
            return a == null ? -1 : 1;
        }
        else if (isComparable(a)) {
            // implements Comparable
            return a.compareTo(b);
        }
        else if (String.isString(a) && String.isString(b)) {
            // native string comparison
            return a.localeCompare(b);
        }
        else if (moment.isMoment(a) && moment.isMoment(b)) {
            return a.valueOf() - b.valueOf();
        }
        else if (moment.isDuration(a) && moment.isDuration(b)) {
            return a.asMilliseconds() - b.asMilliseconds();
        }
        else if (moment.isDate(a) && moment.isDate(b)) {
            return moment(a).valueOf() - moment(b).valueOf();
        }
        else if (Object.isObject(a) || Object.isObject(b)) {
            // if either side is an object then we have failed referencial equality (first compare)
            return -1;
        }
        else {
            // fallback on a basic equality check
            var c = a - b;
            // it's possible that our basic check failed, so default to zero
            return c == null || isNaN(c) ? 0 : c;
        }
    };
    ValueComparer.prototype.compare = function (a, b) {
        return this.comparison(a, b);
    };
    ValueComparer.displayName = 'ValueComparer';
    ValueComparer.Default = new ValueComparer();
    return ValueComparer;
}());
exports.ValueComparer = ValueComparer;
function compare(a, b) {
    return ValueComparer.DefaultComparison(a, b) === 0;
}
exports.compare = compare;
var SortDirection;
(function (SortDirection) {
    SortDirection[SortDirection["Ascending"] = 1] = "Ascending";
    SortDirection[SortDirection["Descending"] = 2] = "Descending";
})(SortDirection = exports.SortDirection || (exports.SortDirection = {}));
var ObjectComparer = /** @class */ (function () {
    function ObjectComparer() {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        var _this = this;
        var comparers = args;
        var defaultSortField = args[0] == null || !String.isNullOrEmpty(args[0])
            ? args.shift()
            : undefined;
        this.comparerMap = {};
        this.comparerMap[ObjectComparer.DefaultComparerKey] = ObjectComparer.createFieldComparer(ObjectComparer.DefaultComparerKey, ValueComparer.DefaultComparison);
        comparers.forEach(function (x) { return (_this.comparerMap[x.field] = x); });
        if (defaultSortField != null) {
            this.defaultComparer = this.getComparer(defaultSortField);
        }
    }
    ObjectComparer.createFieldComparer = function (field, comparison, valueSelector) {
        return {
            field: field,
            compare: comparison,
            valueSelector: valueSelector,
        };
    };
    ObjectComparer.prototype.getComparer = function (field) {
        var comparer = this.comparerMap[field || ObjectComparer.DefaultComparerKey] ||
            this.comparerMap[ObjectComparer.DefaultComparerKey];
        if (String.isNullOrEmpty(comparer.field) && !String.isNullOrEmpty(field)) {
            comparer = Object.assign({}, comparer, { field: field });
        }
        return comparer;
    };
    ObjectComparer.prototype.getCompare = function (comparer) {
        return comparer.compare || ValueComparer.DefaultComparison;
    };
    ObjectComparer.prototype.getValue = function (source, comparer) {
        return comparer.valueSelector == null
            ? source[comparer.field]
            : comparer.valueSelector(source, comparer.field);
    };
    ObjectComparer.prototype.sortIterable = function (source, field, direction) {
        var _this = this;
        if (direction === void 0) { direction = SortDirection.Ascending; }
        var comparer = this.getComparer(field);
        var defaultComparer = this.defaultComparer;
        if (direction === SortDirection.Ascending) {
            var orderedSource = source.orderBy(function (x) { return _this.getValue(x, comparer); }, this.getCompare(comparer));
            source = orderedSource;
            if (defaultComparer != null) {
                source = orderedSource.thenBy(function (x) { return _this.getValue(x, defaultComparer); }, this.getCompare(defaultComparer));
            }
        }
        else if (direction === SortDirection.Descending) {
            var orderedSource = source.orderByDescending(function (x) { return _this.getValue(x, comparer); }, comparer.compare);
            source = orderedSource;
            if (defaultComparer != null) {
                source = orderedSource.thenByDescending(function (x) { return _this.getValue(x, defaultComparer); }, this.getCompare(defaultComparer));
            }
        }
        return source;
    };
    ObjectComparer.displayName = 'ObjectComparer';
    ObjectComparer.DefaultComparerKey = '';
    return ObjectComparer;
}());
exports.ObjectComparer = ObjectComparer;
//# sourceMappingURL=Compare.js.map