Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var moment = require("moment");
var LogLevel_1 = require("../LogLevel");
var Delegate_1 = require("./Delegate");
var ConsoleLogManager = /** @class */ (function (_super) {
    tslib_1.__extends(ConsoleLogManager, _super);
    function ConsoleLogManager(defaultLevel) {
        var _this = _super.call(this, function (manager) {
            return function (logger, level, text, args) {
                return manager.logAction(logger, level, text, args);
            };
        }, defaultLevel) || this;
        _this.sanitizeConsoleObject();
        return _this;
    }
    ConsoleLogManager.prototype.sanitizeConsoleObject = function () {
        // this is an added protection for IE9 to support console.log even if the
        // developer tools are not active.
        if (typeof window !== 'undefined' &&
            window != null &&
            window.console == null) {
            window.console = ConsoleLogManager.dummyConsole;
        }
    };
    ConsoleLogManager.prototype.getColorStyle = function (bgColor, color) {
        if (bgColor === void 0) { bgColor = 'transparent'; }
        if (color === void 0) { color = 'black'; }
        return [
            'display:block',
            'border:0',
            'margin:0',
            'padding:5px 0',
            'line-height:19px',
            "background-color:" + bgColor,
            "color:" + color,
        ].join(';');
    };
    ConsoleLogManager.prototype.getStyles = function (level) {
        var styles = [];
        if (window.chrome != null) {
            if (level >= LogLevel_1.LogLevel.Fatal) {
                // do nothing
            }
            else if (level >= LogLevel_1.LogLevel.Error) {
                // do nothing
            }
            else if (level >= LogLevel_1.LogLevel.Warn) {
                // do nothing
            }
            else if (level >= LogLevel_1.LogLevel.Info) {
                styles.push(this.getColorStyle('lightblue'));
            }
            else if (level >= LogLevel_1.LogLevel.Debug) {
                styles.push(this.getColorStyle('lightcyan'));
            }
            else {
                styles.push(this.getColorStyle());
            }
        }
        return styles;
    };
    ConsoleLogManager.prototype.logAction = function (logger, level, text, args) {
        var _this = this;
        try {
            var styles = this.getStyles(level);
            var style = styles.length > 0 ? '%c' : '';
            var timestamp = moment().format('HH:mm:ss.SSS');
            this.logToConsole.apply(this, tslib_1.__spread([level,
                style + "[" + timestamp + "][" + LogLevel_1.getLevelName(level) + "][" + logger.name + "] " + text], styles));
            if (args != null && args.length > 0) {
                args.filterNull().forEach(function (x) {
                    _this.logToConsole(level, x);
                });
            }
        }
        catch (e) {
            // do nothing
        }
    };
    ConsoleLogManager.prototype.getLogFunction = function (level) {
        var noop = function () {
            return;
        };
        var coerceLogFunction = function (fn) {
            if (fn == null) {
                fn = console.log;
            }
            if (fn == null) {
                return noop;
            }
            if (fn instanceof Function) {
                return fn;
            }
            return function (message) {
                var optionalParams = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    optionalParams[_i - 1] = arguments[_i];
                }
                // try and log the message if we can
                if (!String.isNullOrEmpty(message)) {
                    fn(message);
                }
            };
        };
        if (console == null) {
            return noop;
        }
        else if (level >= LogLevel_1.LogLevel.Error) {
            return coerceLogFunction(console.error);
        }
        else if (level >= LogLevel_1.LogLevel.Warn) {
            return coerceLogFunction(console.warn);
        }
        else if (level >= LogLevel_1.LogLevel.Info) {
            return coerceLogFunction(console.info);
        }
        else if (level >= LogLevel_1.LogLevel.Debug) {
            return coerceLogFunction(console.debug);
        }
        else {
            return coerceLogFunction(console.log);
        }
    };
    ConsoleLogManager.prototype.logToConsole = function (level, message) {
        var formatting = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            formatting[_i - 2] = arguments[_i];
        }
        try {
            var fn = this.getLogFunction(level);
            var consoleRef = console === ConsoleLogManager.dummyConsole ? undefined : console;
            fn.apply(consoleRef, [message].concat(formatting));
        }
        catch (e) {
            // do nothing
        }
    };
    ConsoleLogManager.dummyConsole = {
        log: function () {
            return;
        },
    };
    return ConsoleLogManager;
}(Delegate_1.DelegateLogManager));
exports.ConsoleLogManager = ConsoleLogManager;
exports.Default = new ConsoleLogManager(LogLevel_1.DefaultLogLevel);
//# sourceMappingURL=Console.js.map